@@ -7,7 +7,7 @@ import escapism
 from jupyterhub.proxy import Proxy
 from jupyterhub.utils import exponential_backoff
 from kubernetes_asyncio import client
-from traitlets import Unicode
+from traitlets import Dict, Unicode
 
 from .clients import load_config, shared_client
 from .objects import make_ingress
@@ -78,7 +78,7 @@ class KubeIngressProxy(Proxy):
     KubeIngressProxy is an implementation of the JupyterHub Proxy class that
     JupyterHub can be configured to rely on:
 
-        c.JupyterHub.proxy_class = "kubespawner:KubeIngressProxy"
+        c.JupyterHub.proxy_class = "kubespawner.proxy:KubeIngressProxy"
 
     Like all JupyterHub Proxy implementations, KubeIngressProxy will know
     how to respond to hub requests like `get_all_routes`, `add_route`, and
@@ -126,6 +126,9 @@ class KubeIngressProxy(Proxy):
            ```
     """
 
+    # JupyterHub should not try to start or stop this proxy
+    should_start = False
+
     namespace = Unicode(
         config=True,
         help="""
@@ -159,6 +162,30 @@ class KubeIngressProxy(Proxy):
         """,
     )
 
+    common_labels = Dict(
+        {
+            'app': 'jupyterhub',
+            'heritage': 'jupyterhub',
+        },
+        config=True,
+        help="""
+        Extra kubernetes labels to set on all created objects.
+
+        The keys and values must both be strings that match the kubernetes
+        label key / value constraints.
+
+        See `the Kubernetes documentation <https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/>`__
+        for more info on what labels are and why you might want to use them!
+
+        `{username}`, `{servername}`, `{servicename}`, `{routespec}`, `{hubnamespace}`,
+        `{unescaped_username}`, `{unescaped_servername}`, `{unescaped_servicename}` and `{unescaped_routespec}` will be expanded if
+        found within strings of this configuration.
+
+        Names have to be are escaped to follow the [DNS label
+        standard](https://kubernetes.io/docs/concepts/overview/working-with-objects/names/#dns-label-names).
+        """,
+    )
+
     k8s_api_ssl_ca_cert = Unicode(
         "",
         config=True,
@@ -223,6 +250,58 @@ class KubeIngressProxy(Proxy):
         )
         return safe_name
 
+    def _expand_user_properties(self, template, routespec, data):
+        # Make sure username and servername match the restrictions for DNS labels
+        # Note: '-' is not in safe_chars, as it is being used as escape character
+        safe_chars = set(string.ascii_lowercase + string.digits)
+
+        raw_servername = data.get('server_name') or ''
+        safe_servername = escapism.escape(
+            raw_servername, safe=safe_chars, escape_char='-'
+        ).lower()
+
+        hub_namespace = self._namespace_default()
+        if hub_namespace == "default":
+            hub_namespace = "user"
+
+        raw_username = data.get('user')
+        safe_username = escapism.escape(
+            raw_username, safe=safe_chars, escape_char='-'
+        ).lower()
+
+        raw_servicename = data.get('services')
+        safe_servicename = escapism.escape(
+            raw_servicename, safe=safe_chars, escape_char='-'
+        ).lower()
+
+        raw_routespec = routespec
+        safe_routespec = self._safe_name_for_routespec(routespec)
+
+        rendered = template.format(
+            username=safe_username,
+            unescaped_username=raw_username,
+            servername=safe_servername,
+            unescaped_servername=raw_servername,
+            servicename=safe_servicename,
+            unescaped_servicename=raw_servicename,
+            routespec=safe_routespec,
+            unescaped_routespec=raw_routespec,
+            hubnamespace=hub_namespace,
+        )
+        # strip trailing - delimiter in case of empty servername.
+        # k8s object names cannot have trailing -
+        return rendered.rstrip("-")
+
+    def _expand_all(self, src, routespec, data):
+        if isinstance(src, list):
+            return [self._expand_all(i, routespec, data) for i in src]
+        elif isinstance(src, dict):
+            return {k: self._expand_all(v, routespec, data) for k, v in src.items()}
+        elif isinstance(src, str):
+            return self._expand_user_properties(src, routespec, data)
+        else:
+            return src
+
     async def _delete_if_exists(self, kind, safe_name, future):
         try:
             await future
@@ -238,13 +317,14 @@ class KubeIngressProxy(Proxy):
         # 'data' is JSON encoded and put in an annotation - we don't need to query for it
 
         safe_name = self._safe_name_for_routespec(routespec).lower()
-        labels = {
-            'heritage': 'jupyterhub',
-            'component': self.component_label,
-            'hub.jupyter.org/proxy-route': 'true',
-        }
+        common_labels = self._expand_all(self.common_labels, routespec, data)
+        common_labels.update({'component': self.component_label})
         endpoint, service, ingress = make_ingress(
-            safe_name, routespec, target, labels, data
+            name=safe_name,
+            routespec=routespec,
+            target=target,
+            common_labels=common_labels,
+            data=data,
         )
 
         async def ensure_object(create_func, patch_func, body, kind):
@@ -279,7 +359,7 @@ class KubeIngressProxy(Proxy):
                 'Could not find endpoints/%s after creating it' % safe_name,
             )
         else:
-            delete_endpoint = await self.core_api.delete_namespaced_endpoints(
+            delete_endpoint = self.core_api.delete_namespaced_endpoints(
                 name=safe_name,
                 namespace=self.namespace,
                 body=client.V1DeleteOptions(grace_period_seconds=0),
@@ -321,19 +401,19 @@ class KubeIngressProxy(Proxy):
 
         delete_options = client.V1DeleteOptions(grace_period_seconds=0)
 
-        delete_endpoint = await self.core_api.delete_namespaced_endpoints(
+        delete_endpoint = self.core_api.delete_namespaced_endpoints(
             name=safe_name,
             namespace=self.namespace,
             body=delete_options,
         )
 
-        delete_service = await self.core_api.delete_namespaced_service(
+        delete_service = self.core_api.delete_namespaced_service(
             name=safe_name,
             namespace=self.namespace,
             body=delete_options,
         )
 
-        delete_ingress = await self.networking_api.delete_namespaced_ingress(
+        delete_ingress = self.networking_api.delete_namespaced_ingress(
             name=safe_name,
             namespace=self.namespace,
             body=delete_options,
